@@ -57,6 +57,7 @@
 #include <xercesc/util/XMLUniDefs.hpp>
 
 #ifdef WIN32
+# define WIN32_LEAN_AND_MEAN
 # include <process.h>
 # define XP_WIN32
 #else
@@ -113,6 +114,8 @@ extern "C" NSAPI_PUBLIC void nsapi_shib_exit(void*)
 
 extern "C" NSAPI_PUBLIC int nsapi_shib_init(pblock* pb, ::Session* sn, Request* rq)
 {
+    static char _funcname[] = "nsapi_shib_init";
+
     // Save off a default hostname for this virtual server.
     char* name=pblock_findval("server-name",pb);
     if (name)
@@ -134,12 +137,14 @@ extern "C" NSAPI_PUBLIC int nsapi_shib_init(pblock* pb, ::Session* sn, Request*
         }
     }
 
-    log_error(LOG_INFORM,"nsapi_shib_init",sn,rq,"nsapi_shib loaded for host (%s)",g_ServerName.c_str());
+    static char _logmsg[] = "nsapi_shib loaded for host (%s)";
+    log_error(LOG_INFORM,_funcname,sn,rq,_logmsg,g_ServerName.c_str());
 
     const char* schemadir=pblock_findval("shib-schemas",pb);
     const char* prefix=pblock_findval("shib-prefix",pb);
 
     g_Config=&SPConfig::getConfig();
+    g_Config->deprecation().warn("NSAPI filter");
     g_Config->setFeatures(
         SPConfig::Listener |
         SPConfig::Caching |
@@ -297,16 +302,18 @@ public:
     return pblock_findval("method", m_rq->reqpb);
   }
   string getContentType() const {
+    static char _contenttype[] = "content-type";
     char* content_type = nullptr;
-    if (request_header("content-type", &content_type, m_sn, m_rq) != REQ_PROCEED)
+    if (request_header(_contenttype, &content_type, m_sn, m_rq) != REQ_PROCEED)
         return "";
     return content_type ? content_type : "";
   }
   long getContentLength() const {
+    static char _contentlen[] = "content-length";
     if (m_gotBody)
         return m_body.length();
     char* content_length=nullptr;
-    if (request_header("content-length", &content_length, m_sn, m_rq) != REQ_PROCEED)
+    if (request_header(_contentlen, &content_length, m_sn, m_rq) != REQ_PROCEED)
         return 0;
     return content_length ? atoi(content_length) : 0;
   }
@@ -315,18 +322,20 @@ public:
     return ret.empty() ? pblock_findval("ip", m_sn->client) : ret;
   }
   void log(SPLogLevel level, const string& msg) const {
+    static char _module[] = "nsapi_shib";
     AbstractSPRequest::log(level,msg);
     if (level>=SPError)
-        log_error(LOG_FAILURE, "nsapi_shib", m_sn, m_rq, const_cast<char*>(msg.c_str()));
+        log_error(LOG_FAILURE, _module, m_sn, m_rq, const_cast<char*>(msg.c_str()));
   }
   const char* getQueryString() const {
     return pblock_findval("query", m_rq->reqpb);
   }
   const char* getRequestBody() const {
+    static char _contentlen[] = "content-length";
     if (m_gotBody)
         return m_body.c_str();
     char* content_length=nullptr;
-    if (request_header("content-length", &content_length, m_sn, m_rq) != REQ_PROCEED || !content_length) {
+    if (request_header(_contentlen, &content_length, m_sn, m_rq) != REQ_PROCEED || !content_length) {
         m_gotBody = true;
         return nullptr;
     }
@@ -425,10 +434,13 @@ public:
       param_free(pblock_remove("content-type", m_rq->srvhdrs));
       setResponseHeader("Content-Type", type);
   }
-  void setResponseHeader(const char* name, const char* value) {
-    HTTPResponse::setResponseHeader(name, value);
-    if (name) {
-        pblock_nvinsert(name, value, m_rq->srvhdrs);
+  void setResponseHeader(const char* name, const char* value, bool replace=false) {
+    HTTPResponse::setResponseHeader(name, value, replace);
+    if (name && *name) {
+        if (replace || !value)
+            param_free(pblock_remove(name, m_rq->srvhdrs));
+        if (value && *value)
+            pblock_nvinsert(name, value, m_rq->srvhdrs);
     }
   }
 
@@ -479,10 +491,10 @@ int WriteClientError(::Session* sn, Request* rq, char* func, char* msg)
     return REQ_ABORTED;
 }
 
-#undef FUNC
-#define FUNC "shibboleth"
 extern "C" NSAPI_PUBLIC int nsapi_shib(pblock* pb, ::Session* sn, Request* rq)
 {
+  static char _func[] = "nsapi_shib";
+
   string threadid("[");
   threadid += lexical_cast<string>(getpid()) + "] nsapi_shib";
   xmltooling::NDC ndc(threadid.c_str());
@@ -512,23 +524,25 @@ extern "C" NSAPI_PUBLIC int nsapi_shib(pblock* pb, ::Session* sn, Request* rq)
     // this user is ok.
     return REQ_PROCEED;
   }
-  catch (std::exception& e) {
-    log_error(LOG_FAILURE,FUNC,sn,rq,const_cast<char*>(e.what()));
-    return WriteClientError(sn, rq, FUNC, "Shibboleth module threw an exception, see web server log for error.");
+  catch (const std::exception& e) {
+    log_error(LOG_FAILURE,_func,sn,rq,const_cast<char*>(e.what()));
+    static char _errmsg[] = "Shibboleth module threw an exception, see web server log for error.";
+    return WriteClientError(sn, rq, _func, _errmsg);
   }
   catch (...) {
-    log_error(LOG_FAILURE,FUNC,sn,rq,const_cast<char*>("Shibboleth module threw an unknown exception."));
+    static char _errmsg2[] = "Shibboleth module threw an unknown exception.";
+    log_error(LOG_FAILURE,_func,sn,rq,_errmsg2);
     if (g_catchAll)
-        return WriteClientError(sn, rq, FUNC, "Shibboleth module threw an unknown exception.");
+        return WriteClientError(sn, rq, _func, _errmsg2);
     throw;
   }
 }
 
 
-#undef FUNC
-#define FUNC "shib_handler"
 extern "C" NSAPI_PUBLIC int shib_handler(pblock* pb, ::Session* sn, Request* rq)
 {
+  static char _func[] = "shib_handler";
+
   string threadid("[");
   threadid += lexical_cast<string>(getpid()) + "] shib_handler";
   xmltooling::NDC ndc(threadid.c_str());
@@ -539,16 +553,19 @@ extern "C" NSAPI_PUBLIC int shib_handler(pblock* pb, ::Session* sn, Request* rq)
     pair<bool,long> res = stn.getServiceProvider().doHandler(stn);
     if (res.first) return (int)res.second;
 
-    return WriteClientError(sn, rq, FUNC, "Shibboleth handler did not do anything.");
+    static char _errmsg[] = "Shibboleth handler did not do anything.";
+    return WriteClientError(sn, rq, _func, _errmsg);
   }
-  catch (std::exception& e) {
-    log_error(LOG_FAILURE,FUNC,sn,rq,const_cast<char*>(e.what()));
-    return WriteClientError(sn, rq, FUNC, "Shibboleth handler threw an exception, see web server log for error.");
+  catch (const std::exception& e) {
+    static char _errmsg2[] = "Shibboleth handler threw an exception, see web server log for error.";
+    log_error(LOG_FAILURE,_func,sn,rq,const_cast<char*>(e.what()));
+    return WriteClientError(sn, rq, _func, _errmsg2);
   }
   catch (...) {
-    log_error(LOG_FAILURE,FUNC,sn,rq,"unknown exception caught in Shibboleth handler");
+    static char _errmsg3[] = "Shibboleth handler threw an unknown exception.";
+    log_error(LOG_FAILURE,_func,sn,rq,_errmsg3);
     if (g_catchAll)
-        return WriteClientError(sn, rq, FUNC, "Shibboleth handler threw an unknown exception.");
+        return WriteClientError(sn, rq, _func, _errmsg3);
     throw;
   }
 }
@@ -557,7 +574,7 @@ extern "C" NSAPI_PUBLIC int shib_handler(pblock* pb, ::Session* sn, Request* rq)
 class SunRequestMapper : public virtual RequestMapper, public virtual PropertySet
 {
 public:
-    SunRequestMapper(const xercesc::DOMElement* e);
+    SunRequestMapper(const xercesc::DOMElement* e, bool deprecationSupport=true);
     ~SunRequestMapper() {}
     Lockable* lock() { return m_mapper->lock(); }
     void unlock() { m_stKey->setData(nullptr); m_propsKey->setData(nullptr); m_mapper->unlock(); }
@@ -570,8 +587,7 @@ public:
     pair<bool,const XMLCh*> getXMLString(const char* name, const char* ns=nullptr) const;
     pair<bool,unsigned int> getUnsignedInt(const char* name, const char* ns=nullptr) const;
     pair<bool,int> getInt(const char* name, const char* ns=nullptr) const;
-    void getAll(map<string,const char*>& properties) const;
-    const PropertySet* getPropertySet(const char* name, const char* ns=shibspconstants::ASCII_SHIB2SPCONFIG_NS) const;
+    const PropertySet* getPropertySet(const char* name, const char* ns=shibspconstants::ASCII_SHIBSPCONFIG_NS) const;
     const xercesc::DOMElement* getElement() const;
 
 private:
@@ -579,13 +595,13 @@ private:
     scoped_ptr<ThreadKey> m_stKey, m_propsKey;
 };
 
-RequestMapper* SunRequestMapFactory(const xercesc::DOMElement* const & e)
+RequestMapper* SunRequestMapFactory(const xercesc::DOMElement* const & e, bool deprecationSupport)
 {
     return new SunRequestMapper(e);
 }
 
-SunRequestMapper::SunRequestMapper(const xercesc::DOMElement* e)
-    : m_mapper(SPConfig::getConfig().RequestMapperManager.newPlugin(XML_REQUEST_MAPPER,e)),
+SunRequestMapper::SunRequestMapper(const xercesc::DOMElement* e, bool deprecationSupport)
+    : m_mapper(SPConfig::getConfig().RequestMapperManager.newPlugin(XML_REQUEST_MAPPER,e, deprecationSupport)),
         m_stKey(ThreadKey::create(nullptr)),
         m_propsKey(ThreadKey::create(nullptr))
 {
@@ -606,8 +622,8 @@ pair<bool,bool> SunRequestMapper::getBool(const char* name, const char* ns) cons
     if (stn && !ns && name) {
         // Override boolean properties.
         const char* param=pblock_findval(name,stn->m_pb);
-        if (param && (!strcmp(param,"1") || !strcasecmp(param,"true")))
-            return make_pair(true,true);
+		if (param)
+			return make_pair(true, !strcmp(param, "1") || !strcasecmp(param, "true"));
     }
     return s ? s->getBool(name,ns) : make_pair(false,false);
 }
@@ -667,25 +683,6 @@ pair<bool,int> SunRequestMapper::getInt(const char* name, const char* ns) const
     return s ? s->getInt(name,ns) : pair<bool,int>(false,0);
 }
 
-void SunRequestMapper::getAll(map<string,const char*>& properties) const
-{
-    const ShibTargetNSAPI* stn=reinterpret_cast<const ShibTargetNSAPI*>(m_stKey->getData());
-    const PropertySet* s=reinterpret_cast<const PropertySet*>(m_propsKey->getData());
-    if (s)
-        s->getAll(properties);
-    if (!stn)
-        return;
-    properties["authType"] = "shibboleth";
-    const pb_entry* entry;
-    for (int i=0; i<stn->m_pb->hsize; ++i) {
-        entry = stn->m_pb->ht[i];
-        while (entry) {
-            properties[entry->param->name] = entry->param->value;
-            entry = entry->next;
-        }
-    }
-}
-
 const PropertySet* SunRequestMapper::getPropertySet(const char* name, const char* ns) const
 {
     const PropertySet* s=reinterpret_cast<const PropertySet*>(m_propsKey->getData());
