@@ -32,12 +32,14 @@
 #include "handler/SessionInitiator.h"
 #include "impl/XMLApplication.h"
 
+#define BOOST_BIND_GLOBAL_PLACEHOLDERS
 #include <boost/bind.hpp>
 #include <boost/lexical_cast.hpp>
 #include <boost/algorithm/string.hpp>
 #include <xercesc/util/XMLStringTokenizer.hpp>
 #include <xmltooling/XMLToolingConfig.h>
 #include <xmltooling/util/NDC.h>
+#include <xmltooling/util/XMLConstants.h>
 #include <xmltooling/util/XMLHelper.h>
 
 #ifndef SHIBSP_LITE
@@ -123,7 +125,8 @@ XMLApplication::XMLApplication(
     map<string,string> remapperMap;
     remapperMap[shibspconstants::ASCII_SHIB2SPCONFIG_NS] = shibspconstants::ASCII_SHIB3SPCONFIG_NS;
     remapperMap["relayStateLimit"] = "redirectLimit";
-    remapperMap["relayStateWhitelist"] = "redirectWhitelist";
+    remapperMap["relayStateWhitelist"] = "redirectAllow";
+    remapperMap["redirectWhitelist"] = "redirectAllow";
     DOMPropertySet::STLRemapper remapper(remapperMap);
     load(e, nullptr, this, &remapper);
 
@@ -140,19 +143,35 @@ XMLApplication::XMLApplication(
             else if (!strcmp(prop.second, "host"))
                 m_redirectLimit = REDIRECT_LIMIT_HOST;
             else {
-                if (!strcmp(prop.second, "exact+whitelist"))
-                    m_redirectLimit = REDIRECT_LIMIT_EXACT_WHITELIST;
-                else if (!strcmp(prop.second, "host+whitelist"))
-                    m_redirectLimit = REDIRECT_LIMIT_HOST_WHITELIST;
-                else if (!strcmp(prop.second, "whitelist"))
-                    m_redirectLimit = REDIRECT_LIMIT_WHITELIST;
-                else
+                if (!strcmp(prop.second, "exact+allow")) {
+                    m_redirectLimit = REDIRECT_LIMIT_EXACT_ALLOW;
+                }
+                else if (!strcmp(prop.second, "host+allow")) {
+                    m_redirectLimit = REDIRECT_LIMIT_HOST_ALLOW;
+                }
+                else if (!strcmp(prop.second, "allow")) {
+                    m_redirectLimit = REDIRECT_LIMIT_ALLOW;
+                }
+                else if (!strcmp(prop.second, "exact+whitelist")) {
+                    SPConfig::getConfig().deprecation().warn("redirectLimit of \"exact+whitelist\" replaced by \"exact+allow\"");
+                    m_redirectLimit = REDIRECT_LIMIT_EXACT_ALLOW;
+                }
+                else if (!strcmp(prop.second, "host+whitelist")) {
+                    SPConfig::getConfig().deprecation().warn("redirectLimit of \"host+whitelist\" replaced by \"host+allow\"");
+                    m_redirectLimit = REDIRECT_LIMIT_HOST_ALLOW;
+                }
+                else if (!strcmp(prop.second, "whitelist")) {
+                    SPConfig::getConfig().deprecation().warn("redirectLimit of \"whitelist\" replaced by \"allow\"");
+                    m_redirectLimit = REDIRECT_LIMIT_ALLOW;
+                }
+                else {
                     throw ConfigurationException("Unrecognized redirectLimit setting ($1)", params(1, prop.second));
-                prop = sessionProps->getString("redirectWhitelist");
+                }
+                prop = sessionProps->getString("redirectAllow");
                 if (prop.first) {
                     string dup(prop.second);
                     trim(dup);
-                    split(m_redirectWhitelist, dup, is_space(), algorithm::token_compress_on);
+                    split(m_redirectAllow, dup, is_space(), algorithm::token_compress_on);
                 }
             }
         }
@@ -206,7 +225,7 @@ XMLApplication::XMLApplication(
         m_hash += (DIGITS[0x0F & *ch]);
     }
 
-    doAttributeInfo();
+    doAttributeInfo(log);
 
     if (conf.isEnabled(SPConfig::Handlers))
         doHandlers(pp, e, log);
@@ -229,7 +248,7 @@ XMLApplication::XMLApplication(
 #ifndef SHIBSP_LITE
     nlist = e->getElementsByTagNameNS(samlconstants::SAML20_NS, Audience::LOCAL_NAME);
     if (nlist && nlist->getLength()) {
-        log.warn("DEPRECATED: use of <saml:Audience> elements outside of a Security Policy Rule");
+        SPConfig::getConfig().deprecation().warn("use of <saml:Audience> elements outside of a Security Policy Rule");
         for (XMLSize_t i = 0; i < nlist->getLength(); ++i)
             if (nlist->item(i)->getParentNode()->isSameNode(e) && nlist->item(i)->hasChildNodes())
                 m_audiences.push_back(nlist->item(i)->getFirstChild()->getNodeValue());
@@ -397,12 +416,20 @@ template <class T> T* XMLApplication::doChainedPlugins(
     return nullptr;
 }
 
-void XMLApplication::doAttributeInfo()
+void XMLApplication::doAttributeInfo(Category& log)
 {
     // Populate prefix pair.
     m_attributePrefix.second = "HTTP_";
     pair<bool,const char*> prefix = getString("attributePrefix");
     if (prefix.first) {
+#ifdef HAVE_STRCASECMP
+        if (!strcasecmp(prefix.second, "HTTP_")) {
+#else
+        if (!stricmp(prefix.second, "HTTP_")) {
+#endif
+            log.warn("an attributePrefix of HTTP_ allows header smuggling and is unsafe; this will stop working in a future version");
+        }
+
         m_attributePrefix.first = prefix.second;
         const char* pch = prefix.second;
         while (*pch) {
@@ -551,8 +578,8 @@ void XMLApplication::doHandlers(const ProtocolProvider* pp, const DOMElement* e,
             boost::shared_ptr<Handler> handler;
             if (XMLString::equals(child->getLocalName(), _AssertionConsumerService)) {
                 string bindprop(XMLHelper::getAttrString(child, nullptr, Binding));
-                if (bindprop.empty()) {
-                    log.error("AssertionConsumerService element has no Binding attribute, skipping it...");
+                if (bindprop.empty() || !*child->getAttributeNS(nullptr, Location)) {
+                    log.error("AssertionConsumerService element has empty Binding or Location attribute, skipping it...");
                     child = XMLHelper::getNextSiblingElement(child);
                     continue;
                 }
@@ -617,8 +644,8 @@ void XMLApplication::doHandlers(const ProtocolProvider* pp, const DOMElement* e,
             }
             else if (XMLString::equals(child->getLocalName(), _ArtifactResolutionService)) {
                 string bindprop(XMLHelper::getAttrString(child, nullptr, Binding));
-                if (bindprop.empty()) {
-                    log.error("ArtifactResolutionService element has no Binding attribute, skipping it...");
+                if (bindprop.empty() || !*child->getAttributeNS(nullptr, Location)) {
+                    log.error("ArtifactResolutionService element has empty Binding or Location attribute, skipping it...");
                     child = XMLHelper::getNextSiblingElement(child);
                     continue;
                 }
@@ -640,8 +667,8 @@ void XMLApplication::doHandlers(const ProtocolProvider* pp, const DOMElement* e,
             }
             else if (XMLString::equals(child->getLocalName(), _SingleLogoutService)) {
                 string bindprop(XMLHelper::getAttrString(child, nullptr, Binding));
-                if (bindprop.empty()) {
-                    log.error("SingleLogoutService element has no Binding attribute, skipping it...");
+                if (bindprop.empty() || !*child->getAttributeNS(nullptr, Location)) {
+                    log.error("SingleLogoutService element has empty Binding or Location attribute, skipping it...");
                     child = XMLHelper::getNextSiblingElement(child);
                     continue;
                 }
@@ -651,8 +678,8 @@ void XMLApplication::doHandlers(const ProtocolProvider* pp, const DOMElement* e,
             }
             else if (XMLString::equals(child->getLocalName(), _ManageNameIDService)) {
                 string bindprop(XMLHelper::getAttrString(child, nullptr, Binding));
-                if (bindprop.empty()) {
-                    log.error("ManageNameIDService element has no Binding attribute, skipping it...");
+                if (bindprop.empty() || !*child->getAttributeNS(nullptr, Location)) {
+                    log.error("ManageNameIDService element has empty Binding or Location attribute, skipping it...");
                     child = XMLHelper::getNextSiblingElement(child);
                     continue;
                 }
@@ -716,6 +743,21 @@ void XMLApplication::doSSO(const ProtocolProvider& pp, set<string>& protocols, D
                 if (!hasChildElements) {
                     // Append a session initiator element of the designated type to the root element.
                     DOMElement* sidom = e->getOwnerDocument()->createElementNS(e->getNamespaceURI(), _SessionInitiator);
+
+                    // Copy in any attributes from the <SSO> element so they can be accessed as properties in the SI handler
+                    // but more importantly the MessageEncoders, which are DOM-aware only, not SP property-aware.
+                    // The property-based lookups will walk up the DOM tree but the DOM-only code won't.
+                    for (XMLSize_t p = 0; p < ssopropslen; ++p) {
+                        DOMNode* ssoprop = ssoprops->item(p);
+                        if (ssoprop->getNodeType() == DOMNode::ATTRIBUTE_NODE) {
+                            sidom->setAttributeNS(
+                                ((DOMAttr*)ssoprop)->getNamespaceURI(),
+                                ((DOMAttr*)ssoprop)->getLocalName(),
+                                ((DOMAttr*)ssoprop)->getValue()
+                            );
+                        }
+                    }
+
                     sidom->setAttributeNS(nullptr, _type, inittype.second);
                     e->appendChild(sidom);
                     log.info("adding SessionInitiator of type (%s) to chain (/Login)", initiator->getString("id").second);
@@ -740,7 +782,8 @@ void XMLApplication::doSSO(const ProtocolProvider& pp, set<string>& protocols, D
                 if (idprop.first && pathprop.first) {
                     DOMElement* acsdom = e->getOwnerDocument()->createElementNS(samlconstants::SAML20MD_NS, _AssertionConsumerService);
 
-                    // Copy in any attributes from the <SSO> element so they can be accessed as properties in the ACS handler.
+                    // Copy in any attributes from the <SSO> element so they can be accessed as properties in the ACS handler,
+                    // since the handlers aren't attached to the SSO element.
                     for (XMLSize_t p = 0; p < ssopropslen; ++p) {
                         DOMNode* ssoprop = ssoprops->item(p);
                         if (ssoprop->getNodeType() == DOMNode::ATTRIBUTE_NODE) {
@@ -806,6 +849,21 @@ void XMLApplication::doSSO(const ProtocolProvider& pp, set<string>& protocols, D
             if (discou && *discou) {
                 // Append a session initiator element of the designated type to the root element.
                 DOMElement* sidom = e->getOwnerDocument()->createElementNS(e->getNamespaceURI(), _SessionInitiator);
+
+                // Copy in any attributes from the <SSO> element so they can be accessed as properties in the SI handler
+                // but more importantly the MessageEncoders, which are DOM-aware only, not SP property-aware.
+                // The property-based lookups will walk up the DOM tree but the DOM-only code won't.
+                for (XMLSize_t p = 0; p < ssopropslen; ++p) {
+                    DOMNode* ssoprop = ssoprops->item(p);
+                    if (ssoprop->getNodeType() == DOMNode::ATTRIBUTE_NODE) {
+                        sidom->setAttributeNS(
+                            ((DOMAttr*)ssoprop)->getNamespaceURI(),
+                            ((DOMAttr*)ssoprop)->getLocalName(),
+                            ((DOMAttr*)ssoprop)->getValue()
+                        );
+                    }
+                }
+
                 sidom->setAttributeNS(nullptr, _type, discop);
                 sidom->setAttributeNS(nullptr, _URL, discou);
                 e->appendChild(sidom);
@@ -832,6 +890,7 @@ void XMLApplication::doSSO(const ProtocolProvider& pp, set<string>& protocols, D
         );
     m_handlers.push_back(chain);
     m_sessionInitDefault = chain.get();
+    m_sessionInitMap["Login"] = chain.get();
     m_handlerMap["/Login"] = chain.get();
 }
 
@@ -862,6 +921,21 @@ void XMLApplication::doLogout(const ProtocolProvider& pp, set<string>& protocols
                 if (!hasChildElements) {
                     // Append a logout initiator element of the designated type to the root element.
                     DOMElement* lidom = e->getOwnerDocument()->createElementNS(e->getNamespaceURI(), _LogoutInitiator);
+
+                    // Copy in any attributes from the <Logout> element so they can be accessed as properties in the LI handler
+                    // but more importantly the MessageEncoders, which are DOM-aware only, not SP property-aware.
+                    // The property-based lookups will walk up the DOM tree but the DOM-only code won't.
+                    for (XMLSize_t p = 0; p < slopropslen; ++p) {
+                        DOMNode* sloprop = sloprops->item(p);
+                        if (sloprop->getNodeType() == DOMNode::ATTRIBUTE_NODE) {
+                            lidom->setAttributeNS(
+                                ((DOMAttr*)sloprop)->getNamespaceURI(),
+                                ((DOMAttr*)sloprop)->getLocalName(),
+                                ((DOMAttr*)sloprop)->getValue()
+                            );
+                        }
+                    }
+
                     lidom->setAttributeNS(nullptr, _type, inittype.second);
                     e->appendChild(lidom);
                     log.info("adding LogoutInitiator of type (%s) to chain (/Logout)", initiator->getString("id").second);
@@ -1483,31 +1557,47 @@ void XMLApplication::limitRedirect(const GenericRequest& request, const char* ur
     if (m_redirectLimit == REDIRECT_LIMIT_INHERIT)
         return m_base->limitRedirect(request, url);
     if (m_redirectLimit != REDIRECT_LIMIT_NONE) {
-        vector<string> whitelist;
-        if (m_redirectLimit == REDIRECT_LIMIT_EXACT || m_redirectLimit == REDIRECT_LIMIT_EXACT_WHITELIST) {
-            // Scheme and hostname have to match.
-            if (request.isDefaultPort()) {
-                whitelist.push_back(string(request.getScheme()) + "://" + request.getHostname() + '/');
-            }
-            whitelist.push_back(string(request.getScheme()) + "://" + request.getHostname() + ':' + lexical_cast<string>(request.getPort()) + '/');
-        }
-        else if (m_redirectLimit == REDIRECT_LIMIT_HOST || m_redirectLimit == REDIRECT_LIMIT_HOST_WHITELIST) {
-            // Allow any scheme or port.
-            whitelist.push_back(string("https://") + request.getHostname() + '/');
-            whitelist.push_back(string("http://") + request.getHostname() + '/');
-            whitelist.push_back(string("https://") + request.getHostname() + ':');
-            whitelist.push_back(string("http://") + request.getHostname() + ':');
+
+        // This is ugly, but the purpose is to prevent blocking legitimate redirects
+        // that lack a trailing slash after the hostname. If there are fewer than 3
+        // slashes, we assume the hostname wasn't terminated.
+        string urlcopy(url);
+        if (count(urlcopy.begin(), urlcopy.end(), '/') < 3) {
+            urlcopy += '/';
         }
 
-        static bool (*startsWithI)(const char*,const char*) = XMLString::startsWithI;
-        if (!whitelist.empty() && find_if(whitelist.begin(), whitelist.end(),
-                boost::bind(startsWithI, url, boost::bind(&string::c_str, _1))) != whitelist.end()) {
-            return;
+        vector<string> allowlist;
+        if (m_redirectLimit == REDIRECT_LIMIT_EXACT || m_redirectLimit == REDIRECT_LIMIT_EXACT_ALLOW) {
+            // Scheme and hostname have to match.
+            if (request.isDefaultPort()) {
+                allowlist.push_back(string(request.getScheme()) + "://" + request.getHostname() + '/');
+            }
+            allowlist.push_back(string(request.getScheme()) + "://" + request.getHostname() + ':' + lexical_cast<string>(request.getPort()) + '/');
         }
-        else if (!m_redirectWhitelist.empty() && find_if(m_redirectWhitelist.begin(), m_redirectWhitelist.end(),
-                boost::bind(startsWithI, url, boost::bind(&string::c_str, _1))) != m_redirectWhitelist.end()) {
-            return;
+        else if (m_redirectLimit == REDIRECT_LIMIT_HOST || m_redirectLimit == REDIRECT_LIMIT_HOST_ALLOW) {
+            // Allow any scheme or port.
+            allowlist.push_back(string("https://") + request.getHostname() + '/');
+            allowlist.push_back(string("http://") + request.getHostname() + '/');
+            allowlist.push_back(string("https://") + request.getHostname() + ':');
+            allowlist.push_back(string("http://") + request.getHostname() + ':');
         }
+
+        if (!allowlist.empty()) {
+            for (vector<string>::const_iterator i = allowlist.begin(); i != allowlist.end(); ++i) {
+                if (istarts_with(urlcopy, *i)) {
+                    return;
+                }
+            }
+        }
+
+        if (!m_redirectAllow.empty()) {
+            for (vector<string>::const_iterator i = m_redirectAllow.begin(); i != m_redirectAllow.end(); ++i) {
+                if (istarts_with(urlcopy, *i)) {
+                    return;
+                }
+            }
+        }
+
         Category::getInstance(SHIBSP_LOGCAT ".Application").warn("redirectLimit policy enforced, blocked redirect to (%s)", url);
         throw opensaml::SecurityPolicyException("Blocked unacceptable redirect location.");
     }
